@@ -31,11 +31,15 @@ pygtk.require('2.0')
 import gtk
 import gtkhtml2
 import os
+import urllib2
 from portrait import FremantleRotation
 import constants
 from feedmanager import FeedManager
 # For unescape()
 import re, htmlentitydefs
+from threading import Thread
+import time
+
 _ = gettext.gettext
 
 manager = FeedManager()
@@ -80,6 +84,7 @@ class FeedsView(hildon.GtkTreeView):
     FEEDS_ICON_COLUMN = 0
     FEEDS_NAME_COLUMN = 1
     FEEDS_DATA_COLUMN = 2
+    FEEDS_READ_COLUMN = 3
 
     DUMMY_FEED_STATUS = -1
 
@@ -95,24 +100,29 @@ class FeedsView(hildon.GtkTreeView):
         column = gtk.TreeViewColumn('Icon', pix_renderer, pixbuf = 0)
         self.append_column(column);
 
-        column = gtk.TreeViewColumn('Name', text_renderer, text = self.FEEDS_NAME_COLUMN)
-        column.set_cell_data_func(text_renderer, self._view_data_func)
-
+        column = gtk.TreeViewColumn('Name', text_renderer, markup = self.FEEDS_NAME_COLUMN)
+        column.set_expand(True)
         self.append_column(column);
 
-    def _view_data_func(self, column, renderer, model, iter):
-        name = model.get_value(iter, self.FEEDS_NAME_COLUMN)
-        feed_data = model.get_value(iter, self.FEEDS_DATA_COLUMN)
+        date_renderer = gtk.CellRendererText()
+        column = gtk.TreeViewColumn('Date', date_renderer)
+        column.set_cell_data_func(date_renderer, self._unread_view_data_func)
+        self.append_column(column);
 
-        if 'link' in  feed_data.feed:
-            url = feed_data.feed.link
+
+    def _unread_view_data_func(self, column, renderer, model, feed_iter):
+        feed_data, unread_count = model.get(feed_iter, self.FEEDS_DATA_COLUMN,
+                                      self.FEEDS_READ_COLUMN)
+
+        if feed_data.status == self.DUMMY_FEED_STATUS:
+            renderer.set_property('text', '')
+            return
+
+        if unread_count:
+            renderer.set_property('markup', '<span size="xx-small">%d '
+                                  '%s\n%s</span>' % (unread_count, _('unread'), _('items')))
         else:
-            url = feed_data.href
-
-        renderer.set_property('markup',
-                              '%s\n<span size="x-small" style="italic">'
-                              '%s</span>' % \
-                              (name, url))
+            renderer.set_property('markup', '')
 
     def add_dummy_feed(self, url):
 
@@ -123,7 +133,7 @@ class FeedsView(hildon.GtkTreeView):
         dummy_feed_data.href = url
         dummy_feed_data.feed = feedparser.FeedParserDict()
         dummy_feed_data.feed.title = _('Retrieving info...')
-        store.set(feed_iter, 0, None, 1, dummy_feed_data.feed.title, 2, dummy_feed_data)
+        store.set(feed_iter, 0, None, 1, dummy_feed_data.feed.title, 2, dummy_feed_data, 3, -1)
 
         return (dummy_feed_data,feed_iter)
 
@@ -200,7 +210,7 @@ class FeedsWindow(hildon.StackableWindow):
 
         # Load Feeds (could be none)
         try:
-            manager.load()
+            manager.load(self._feed_data_loaded)
         except IOError:
             pass
 
@@ -219,7 +229,7 @@ class FeedsWindow(hildon.StackableWindow):
 
         # Feeds
         self.view = FeedsView(gtk.HILDON_UI_MODE_NORMAL)
-        self.view.set_model (gtk.ListStore(gtk.gdk.Pixbuf, str, gobject.TYPE_PYOBJECT))
+        self.view.set_model (gtk.ListStore(gtk.gdk.Pixbuf, str, gobject.TYPE_PYOBJECT, int))
         self.view.connect ("row-activated", self._on_feed_activated)
         self.view.show()
 
@@ -253,10 +263,23 @@ class FeedsWindow(hildon.StackableWindow):
         hildon.Program.get_instance().add_window(news_window)
         news_window.show()
 
+        # When the news window is destroyed reset the unread cache
+        news_window.connect('delete-event', self._on_news_window_destroyed,
+                            gtk.TreeRowReference(self.view.get_model(), path))
+
+    def _on_news_window_destroyed(self, news_window, event, row_reference):
+        feed_iter = self.view.get_model().get_iter(row_reference.get_path())
+        feed_data = self.view.get_model().get_value(feed_iter, self.view.FEEDS_DATA_COLUMN)
+
+        # Reset unread cache (the value have most likely changed).
+        unread_count = len([entry for entry in feed_data.entries if entry.read == False])
+        self.view.get_model().set(feed_iter, self.view.FEEDS_READ_COLUMN, unread_count)
+
     def _on_exit(self, window, event):
         # Save the feeds status
-        manager.save()
+        manager.save(self._on_feed_data_saved)
 
+    def _on_feed_data_saved(self, user_data):
         # Close application
         gtk.main_quit()
 
@@ -372,6 +395,17 @@ class FeedsWindow(hildon.StackableWindow):
 
             dialog.destroy()
 
+    def _get_feed_title_markup(self, feed_data):
+        if 'link' in  feed_data.feed:
+            url = feed_data.feed.link
+        else:
+            url = feed_data.href
+
+        color = self.style.lookup_color(constants.SECONDARY_TEXT_COLOR).to_string()
+        return '%s\n<span face="monospace" foreground="%s"' \
+            'size="x-small">%s</span>' % \
+            (feed_data.feed.title, color, url)
+
     def _feed_added_cb(self, pixbuf, new_feed_data, row_reference):
 
         # Remove progress information
@@ -389,7 +423,8 @@ class FeedsWindow(hildon.StackableWindow):
             return
 
         # Update model
-        store.set(feed_iter, 0, pixbuf, 1, new_feed_data.feed.title, 2, new_feed_data)
+        store.set(feed_iter, 0, pixbuf, 1, self._get_feed_title_markup(new_feed_data),
+                  2, new_feed_data, 3, len(new_feed_data.entries))
 
         # Take screenshot
         hildon.hildon_gtk_window_take_screenshot(self, False)
@@ -401,7 +436,7 @@ class FeedsWindow(hildon.StackableWindow):
             print 'save preferences'
             dialog.destroy()
 
-    def load_data(self):
+    def _feed_data_loaded(self, user_data):
         # TODO: migrate to "with" statement when available
         try:
             feed_data_list = manager.get_feed_list()
@@ -413,7 +448,12 @@ class FeedsWindow(hildon.StackableWindow):
             for feed_data in feed_data_list:
                 iter = store.append()
                 pixbuf = manager.get_favicon(feed_data.feed.link or feed_data.href)
-                store.set(iter, 0, pixbuf, 1, feed_data.feed.title, 2, feed_data)
+                unread_count = len([entry for entry in feed_data.entries if entry.read == False])
+                store.set(iter, 0, pixbuf, 1, self._get_feed_title_markup(feed_data),
+                          2, feed_data, 3, unread_count)
+                # Improve responsiveness
+#                while gtk.events_pending():
+#                    gtk.main_iteration_do()
 
         # If there is no data then show the new feed dialog
         if len(feed_data_list) == 0:
@@ -426,22 +466,28 @@ class EntriesView(hildon.GtkTreeView):
     ENTRIES_ICON_COLUMN = 0
     ENTRIES_NAME_COLUMN = 1
     ENTRIES_DATA_COLUMN = 2
+    ENTRIES_DATE_COLUMN = 3
 
     def __init__(self, feed_title, ui_mode):
         super(EntriesView, self).__init__(ui_mode)
 
         self.fallback_author = feed_title
-        text_renderer = gtk.CellRendererText()
-        text_renderer.set_property('ellipsize', pango.ELLIPSIZE_END)
-
-        pix_renderer = gtk.CellRendererPixbuf()
 
         # Add columns
+        pix_renderer = gtk.CellRendererPixbuf()
         column = gtk.TreeViewColumn('Icon', pix_renderer, pixbuf = 0)
         self.append_column(column);
 
-        column = gtk.TreeViewColumn('Name', text_renderer, text = self.ENTRIES_NAME_COLUMN)
+        text_renderer = gtk.CellRendererText()
+        text_renderer.set_property('ellipsize', pango.ELLIPSIZE_END)
+        column = gtk.TreeViewColumn('Name', text_renderer)
         column.set_cell_data_func(text_renderer, self._view_data_func)
+        column.set_expand(True)
+        self.append_column(column);
+
+        date_renderer = gtk.CellRendererText()
+        column = gtk.TreeViewColumn('Date', date_renderer, markup = self.ENTRIES_DATE_COLUMN)
+        column.set_expand(False)
         self.append_column(column);
 
     def _view_data_func(self, column, renderer, model, iter):
@@ -465,8 +511,31 @@ class EntriesView(hildon.GtkTreeView):
 
         renderer.set_property('markup',
                               '<span foreground="%s">%s</span>\n<span size="small">%s</span>'
-                              % (color, author, title))
+                              % (color,
+                                 gobject.markup_escape_text(author),
+                                 gobject.markup_escape_text(title)))
 
+    def get_visual_entry_date(self, entry):
+
+        now = time.localtime()
+
+        # Today
+        if now.tm_yday == entry.updated_parsed.tm_yday:
+            return '<span size="xx-small">%s</span>' % \
+                time.strftime('%X', entry.updated_parsed)
+        # This week
+        elif now.tm_yday - entry.updated_parsed.tm_yday < 7 and \
+                entry.updated_parsed.tm_wday < now.tm_wday:
+            return '<span size="xx-small">%s</span>' % \
+                time.strftime('%A', entry.updated_parsed)
+        # This year
+        elif now.tm_year == entry.updated_parsed.tm_year:
+            return '<span size="xx-small">%s</span>' % \
+                time.strftime('%d %B', entry.updated_parsed)
+        # Anything else
+        else:
+            return '<span size="xx-small">%s</span>' % \
+                time.strftime('%x', entry.updated_parsed)
 
 class EntriesWindow(hildon.StackableWindow):
 
@@ -478,7 +547,7 @@ class EntriesWindow(hildon.StackableWindow):
 
         # Feeds
         self.view = EntriesView(feed_data.feed.title, gtk.HILDON_UI_MODE_NORMAL)
-        self.view.set_model (gtk.ListStore(gtk.gdk.Pixbuf, str, gobject.TYPE_PYOBJECT))
+        self.view.set_model (gtk.ListStore(gtk.gdk.Pixbuf, str, gobject.TYPE_PYOBJECT, str))
         self.view.connect ("row-activated", self._on_entry_activated)
         self.view.show()
 
@@ -486,7 +555,8 @@ class EntriesWindow(hildon.StackableWindow):
 
         for entry in self.feed_data.entries:
             entry_iter = model.append()
-            model.set(entry_iter, 1, entry.title, 2, entry)
+            model.set(entry_iter, 1, entry.title, 2, entry,
+                      3, self.view.get_visual_entry_date(entry))
 
         align = gtk.Alignment(0, 0, 1, 1)
         align.set_padding(4, 0 , 16, 16)
@@ -508,12 +578,20 @@ class EntriesWindow(hildon.StackableWindow):
         button = hildon.GtkButton(gtk.HILDON_SIZE_FINGER_HEIGHT)
         button.set_label(_('Update Feed'))
         button.connect('clicked', self._update_feed_cb)
+        menu.append(button)
 
+        button = hildon.GtkButton(gtk.HILDON_SIZE_FINGER_HEIGHT)
+        button.set_label(_('Mark all as Read'))
+        button.connect('clicked', self._mark_all_read_cb)
         menu.append(button)
 
         menu.show_all()
         self.set_app_menu(menu)
 
+    def _mark_all_read_cb(self, button):
+        for entry in self.feed_data.entries:
+            entry['read'] = True
+
     def _on_entry_activated(self, treeview, path, column):
         item_window = ItemWindow(self.feed_data, treeview.get_model(), path)
         hildon.Program.get_instance().add_window(item_window)
@@ -538,7 +616,8 @@ class EntriesWindow(hildon.StackableWindow):
         model.clear()
         for entry in self.feed_data.entries:
             entry_iter = model.append()
-            model.set(entry_iter, 1, entry.title, 2, entry)
+            model.set(entry_iter, 1, entry.title, 2, entry,
+                      3, self.view.get_visual_entry_date(entry))
 
         return
 
@@ -620,6 +699,7 @@ class ItemWindow(hildon.StackableWindow):
         self.view = gtkhtml2.View()
         self.document = gtkhtml2.Document()
         self.view.set_document(self.document)
+        self.document.connect('request-url', self._on_request_url)
         self.set_item(item)
 
         # Alignment for padding
@@ -644,18 +724,71 @@ class ItemWindow(hildon.StackableWindow):
         self.add(align)
         align.show()
 
-#        self._create_menu()
+        self._create_menu()
+
+    def _on_request_url(self, object, url, stream):
+        if (url.lower().startswith("http")):
+            request_thread = Thread(target=self._request_url_in_thread,
+                                    args=(url, stream))
+            request_thread.start()
+
+    def _request_url_in_thread(self, url, stream):
+        f = urllib2.urlopen(url)
+
+        gtk.gdk.threads_enter()
+        stream.write(f.read())
+        stream.close()
+        gtk.gdk.threads_leave()
 
     def _create_menu(self):
         menu = hildon.AppMenu()
 
         button = hildon.GtkButton(gtk.HILDON_SIZE_FINGER_HEIGHT)
-        button.set_label(_('Item Window Empty'))
+        button.set_label(_('Item Details'))
+        button.connect('clicked', self._item_details_cb)
         menu.append(button)
 
         menu.show_all()
         self.set_app_menu(menu)
 
+
+    def _add_to_table (self, table, color, key, value, y_coord):
+        label = gtk.Label('')
+        label.set_markup('<span foreground="%s">%s:</span>' % (color, key))
+        label.set_alignment(1, 0)
+        table.attach(label, 0, 1, y_coord, y_coord + 1)
+
+        value_label = gtk.Label('')
+        value_label.set_markup(value)
+        value_label.set_alignment(0, 0)
+        table.attach(gtk.Label(value), 1, 2, y_coord, y_coord + 1)
+
+    def _item_details_cb(self, button):
+        dialog = gtk.Dialog()
+        dialog.set_title(_('Item Details'))
+
+        pannable = hildon.PannableArea()
+        pannable.set_size_request_policy(hildon.SIZE_REQUEST_CHILDREN)
+
+        item = self.item_window_header.item
+        color = self.style.lookup_color(constants.SECONDARY_TEXT_COLOR).to_string()
+
+        row = 0;
+        table = gtk.Table()
+        table.set_col_spacing(0, 12)
+        self._add_to_table (table, color, _('Title'), item.title, row)
+        row += 1
+        self._add_to_table (table, color, _('Published'),
+                            time.strftime('%x', item.updated_parsed), row)
+
+        pannable.add_with_viewport(table)
+
+        table.show_all()
+        pannable.show()
+        dialog.vbox.add(pannable)
+
+        dialog.show()
+
     def _up_button_clicked(self, button):
 
         # http://faq.pygtk.org/index.py?req=show&file=faq13.051.htp
